#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['ludology']

package_data = \
{'': ['*']}

install_requires = \
['numpy', 'scipy']

extras_require = \
{'dev': ['codecov',
         'hypothesis',
         'ipython[nbconvert]',
         'pytest',
         'pytest-cov',
         'sphinx'],
 'plotting': ['matplotlib']}

setup(name='ludology',
      version='0.1.1',
      description='Ludology is a package for the study of combinatorial game theory.',
      author='Ryan G. James',
      author_email='ryangregoryjames@gmail.com',
      url='https://github.com/autoplectic/ludology',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
     )
